"""
:Created: 6 December 2014
:Author: Lucas Connors

"""

from django import forms
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from jsonfield.fields import JSONField, JSONFormField, JSONWidget

from .conditions import CompareCondition
from .exceptions import InvalidConditionError
from .lists import CondList


__all__ = ['ConditionsWidget', 'ConditionsFormField', 'ConditionsField']


class ConditionsWidget(JSONWidget):

    template_name = 'conditions/conditions_widget.html'

    def __init__(self, *args, **kwargs):
        self.condition_definitions = kwargs.pop('condition_definitions', {})
        if 'attrs' not in kwargs:
            kwargs['attrs'] = {}
        if 'cols' not in kwargs['attrs']:
            kwargs['attrs']['cols'] = 50
        super(ConditionsWidget, self).__init__(*args, **kwargs)

    def render(self, name, value, attrs=None):
        if isinstance(value, CondList):
            value = value.encode()
        textarea = super(ConditionsWidget, self).render(name, value, attrs)

        condition_groups = []
        for groupname, group in self.condition_definitions.iteritems():
            conditions_in_group = []
            for condstr, condition in group.items():
                conditions_in_group.append({
                    'condstr': condstr,
                    'key_required': 'true' if condition.key_required() else 'false',
                    'keys_allowed': condition.keys_allowed,
                    'key_example': condition.key_example(),
                    'operator_required': 'true' if issubclass(condition, CompareCondition) else 'false',
                    'operators': condition.operators().keys() if issubclass(condition, CompareCondition) else [],
                    'operand_example': condition.operand_example() if issubclass(condition, CompareCondition) else '',
                    'help_text': condition.help_text(),
                    'description': condition.full_description(),
                })
            conditions_in_group = sorted(conditions_in_group, key=lambda x: x['condstr'])

            condition_groups.append({
                'groupname': groupname,
                'conditions': conditions_in_group,
            })
        condition_groups = sorted(condition_groups, key=lambda x: x['groupname'])

        context = {
            'textarea': textarea,
            'condition_groups': condition_groups,
        }

        return mark_safe(render_to_string(self.template_name, context))


class ConditionsFormField(JSONFormField):

    def __init__(self, *args, **kwargs):
        self.condition_definitions = kwargs.pop('condition_definitions', {})
        if 'widget' not in kwargs:
            kwargs['widget'] = ConditionsWidget(condition_definitions=self.condition_definitions)
        super(ConditionsFormField, self).__init__(*args, **kwargs)

    def clean(self, value):
        """ Validate conditions by decoding result """
        cleaned_json = super(ConditionsFormField, self).clean(value)
        if cleaned_json is None:
            return

        try:
            CondList.decode(cleaned_json, definitions=self.condition_definitions)
        except InvalidConditionError as e:
            raise forms.ValidationError("Invalid conditions JSON: {error}".format(error=str(e)))
        else:
            return cleaned_json


class ConditionsField(JSONField):
    """
    ConditionsField stores information on when the "value" of the
    instance should be considered True.
    """

    def __init__(self, *args, **kwargs):
        self.condition_definitions = kwargs.pop('definitions', {})
        super(ConditionsField, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        kwargs['condition_definitions'] = self.condition_definitions
        return ConditionsFormField(**kwargs)

    def pre_init(self, value, obj):
        value = super(ConditionsField, self).pre_init(value, obj)
        if isinstance(value, dict):
            value = CondList.decode(value, definitions=self.condition_definitions)
        return value

    def dumps_for_display(self, value):
        if isinstance(value, CondList):
            value = value.encode()
        return super(ConditionsField, self).dumps_for_display(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        if isinstance(value, CondList):
            value = value.encode()
        return super(ConditionsField, self).get_db_prep_value(value, connection, prepared)



"""
:Created: 6 December 2014
:Author: Lucas Connors

"""

__all__ = ['UndefinedConditionError', 'InvalidConditionError']


class UndefinedConditionError(NotImplementedError):
    pass


class InvalidConditionError(ValueError):
    pass



from django.contrib.auth.models import User
from django.db import models

from ..fields import ConditionsField
from ..types import conditions_from_module
from . import conditions


class UserProfile(models.Model):

    user = models.OneToOneField(User)
    level = models.IntegerField(default=1)


class Campaign(models.Model):

    text = models.TextField()
    conditions = ConditionsField(definitions=conditions_from_module(conditions))



"""
:Created: 7 December 2014
:Author: Lucas Connors

"""

import inspect

from .conditions import Condition


__all__ = ['conditions_from_module']


def _iscondition(cond):
    return (
        inspect.isclass(cond)
        and issubclass(cond, Condition)
        and cond.__name__ not in ['Condition', 'CompareCondition']
    )


def conditions_from_module(module):
    all_conditions = {}
    for name, condition in inspect.getmembers(module):
        if _iscondition(condition):
            condition_module = condition.module_name()
            if condition_module not in all_conditions:
                all_conditions[condition_module] = {}
            all_conditions[condition_module][condition.condstr] = condition

    return all_conditions



"""
:Created: 6 December 2014
:Author: Lucas Connors

"""

import abc

from .conditions import Condition
from .exceptions import InvalidConditionError


__all__ = ['CondList', 'CondAllList', 'CondAnyList', 'eval_conditions']


class CondList(list):
    """
    A list of conditions;
    Subclassed to decide how the conditions are evaluated
    """

    __metaclass__ = abc.ABCMeta

    @classmethod
    def decode_item(cls, value, definitions=None):
        if isinstance(value, dict):
            return cls.decode(value, definitions=definitions)
        else:
            return Condition.decode(value, definitions=definitions)

    @classmethod
    def decode_list(cls, value, definitions=None):
        # Validate condlist
        if not isinstance(value, list):
            raise InvalidConditionError("Condition list is invalid.")

        return [cls.decode_item(item, definitions=definitions) for item in value]

    @classmethod
    def decode(cls, value, definitions=None):
        # Validate conddict
        if value is not None and (not isinstance(value, dict) or 'all' in value == 'any' in value):
            raise InvalidConditionError("Conditions dict is invalid. Exactly one of 'all' or 'any' must be used.")

        if 'all' in value:
            return CondAllList(cls.decode_list(value['all'], definitions=definitions))
        elif 'any' in value:
            return CondAnyList(cls.decode_list(value['any'], definitions=definitions))

    @abc.abstractmethod
    def encode(self):
        pass

    @abc.abstractmethod
    def eval(self, user, **kwargs):
        pass


class CondAllList(CondList):
    """
    All conditions must evaluate to True
    """

    def __repr__(self):
        return "<All: {}>".format(super(CondAllList, self).__repr__())

    def encode(self):
        return {
            'all': map(lambda x: x.encode(), self),
        }

    def eval(self, user, **kwargs):
        for condition in self:
            if not condition.eval(user, **kwargs):
                return False
        return True


class CondAnyList(CondList):
    """
    Any conditions may evaluate to True
    """

    def __repr__(self):
        return "<Any: {}>".format(super(CondAnyList, self).__repr__())

    def encode(self):
        return {
            'any': map(lambda x: x.encode(), self),
        }

    def eval(self, user, **kwargs):
        for condition in self:
            if condition.eval(user, **kwargs):
                return True
        return False


def eval_conditions(model, field_name, user, **kwargs):
    conditions = getattr(model, field_name)

    if conditions is None:
        return False

    if isinstance(conditions, dict):
        condition_definitions = model._meta.get_field(field_name).condition_definitions
        conditions = CondList.decode(conditions, definitions=condition_definitions)

    return conditions.eval(user, **kwargs)



from .conditions import Condition, CompareCondition
from .exceptions import UndefinedConditionError, InvalidConditionError
from .fields import ConditionsWidget, ConditionsFormField, ConditionsField
from .lists import CondList, CondAllList, CondAnyList, eval_conditions
from .types import conditions_from_module


__all__ = [
    'Condition', 'CompareCondition', 'UndefinedConditionError', 'InvalidConditionError', 'ConditionsWidget',
    'ConditionsFormField', 'ConditionsField', 'CondList', 'CondAllList', 'CondAnyList', 'eval_conditions',
    'conditions_from_module',
]



#!/usr/bin/env python
import os
from setuptools import setup, find_packages, Command


with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


class TestCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import django
        from django.conf import settings
        from django.core.management import call_command

        settings.configure(
            DATABASES={
                'default': {
                    'NAME': ':memory:',
                    'ENGINE': 'django.db.backends.sqlite3',
                }
            },
            INSTALLED_APPS=(
                'django.contrib.auth',
                'django.contrib.contenttypes',
                'conditions',
            ),
            MIDDLEWARE_CLASSES=()
        )
        django.setup()
        call_command('test')


requirements = [
    'Django >= 1.8, < 1.10a0',
    'django-jsonfield >= 1.0.1, < 2.0.0a0',
]


setup(
    name='django-conditions',
    version='0.9.10',
    packages=find_packages(),
    include_package_data=True,
    license='ISC License',
    description='A Django app that allows creation of conditional logic in admin.',
    long_description=README,
    url='https://github.com/RevolutionTech/django-conditions/',
    author='Lucas Connors',
    author_email='lucas@revolutiontech.ca',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    package_data={
        'conditions': [
            'static/conditions/js/conditions_widget.js',
            'static/conditions/img/icon-information.png',
            'static/conditions/img/icon-information.png',
            'static/conditions/img/icon-information.png',
            'templates/conditions/conditions_widget.html',
        ],
    },
    install_requires=requirements,
    tests_require=requirements,
    cmdclass={'test': TestCommand},
)
